/*
 * icon.cc - code for the Icon widget
 *
 * This file is part of the source distribution of 
 * Jafm - an icon based X11 desktop program 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <iostream.h>

#include <qpainter.h>
#include <qpixmap.h>
#include <qpmcache.h>
#include <qfileinf.h>
#include <qbitmap.h>
#include <qtstream.h>
#include <qregexp.h>
#include <qlist.h>

#include "icon.hh"
#include "iconview.hh"

#include <X11/X.h>
#include <unistd.h>

QList<QString> *Icon::rule_list = 0 ;
extern class Sfm *qApp ;
Icon::Icon (const char *name, const char *zicname=0, int s=10,bool withfile=true)
{
  /* constuct Icon.
   * 
   * if name is nonzero, it refers to a file
   */

  
  if (name) {
    if (withfile) {
      QFile file (name) ;  // assumes we are in right dir?
      finfo = new QFileInfo (file) ;
      CHECK_PTR (finfo) ;
    } else {
      finfo = 0 ;
    }
    label = new char[strlen (name) + 1] ;
    CHECK_PTR (label) ;
    strcpy (label, name) ;
    if (is_drag_droppable()) 
      set_accepts_drag_drop (true) ;
  } else {
    label = 0 ;
    finfo = 0 ;
  }
  
  // has a zicon been suggested?
  if (zicname) { // 
    int n = strlen (zicname) ;
    icon_file = new char[n+1] ;
    CHECK_PTR (icon_file) ;
    strcpy (icon_file, zicname) ;
    program = new IconProgram (zicname) ;
    CHECK_PTR (program) ;
  } else {
    icon_file = suggest_zic (name) ; // always returns something useful
    program = new IconProgram (icon_file) ;
    CHECK_PTR (program) ;
  }

  label_rect = new QRect () ;
  CHECK_PTR (label_rect) ;

  scale = s ; rescale (s) ;

  focus = false ;
  drag_drop = false ;
  selected = false ;
  embossed = false ;

  connect (this, SIGNAL (open_dir_window (const char *)),
	   (const QObject *)qApp, SLOT (open (const char *))) ;

  show () ;
}

// copy constructor (shallow)

Icon::Icon (const Icon * ic)
{

  icon_file = ic->icon_file ;
  program = ic->program ;

  label_rect = new QRect (ic->get_label_rect()) ;
  CHECK_PTR (label_rect) ;
  bound_rect = ic->rect() ;

  if (ic->has_file()) {
    finfo = ic->finfo ;
  } else {
    finfo = 0 ;
  }
  
  label = new char[strlen (ic->label) + 1] ;
  CHECK_PTR (label) ;
  strcpy (label, ic->label) ;
  if (ic->is_drag_droppable()) 
    set_accepts_drag_drop (true) ;
  
  strcpy (pm_normal, ic->pm_normal) ;
  strcpy (pm_selected, ic->pm_selected) ;
  strcpy (pm_focus, ic->pm_focus) ;
  strcpy (pm_drag_drop, ic->pm_drag_drop) ;
  strcpy (pm_embossed, ic->pm_embossed) ;
  strcpy (mask, ic->mask) ;
  scale = ic->scale ;

  focus = false ;
  drag_drop = false ;
  selected = false ;
  embossed = false ;
  show () ;
}


Icon::~Icon ()
{
}

void Icon::draw (QPainter &p)
{
  /* this routine draws the icon
   * on the supplied painter
   * in our position and size
   */
  
  if (hidden) return ;
  
  // work out which pixmap is needed.
  // when necessary, rebuild the pixmap

  QPixmap *pm ;

  if (is_embossed()) {
    pm = get_embossed () ;
  } else {
    if (is_drag_drop()) {
      pm = get_drag_drop () ;
    } else {
      if (is_selected()) {
	pm = get_selected() ;
      } else {
	if (has_focus()) {
	  pm = get_focus() ;
	} else {
	  pm = get_normal () ;
	}
      }
    }
  }
  
  // make sure the mask has not been out-cached!
  // (subtle bug: what if the cache is not big enough to hold two pixmaps?)
  (void) get_mask() ;

  p.drawPixmap (x(), y(), *pm) ;
  
  // do we have a label? if so draw
  if (label) {
    p.setFont (QFont ("Helvetica", scale/8+5 )) ;
    QFontMetrics fm = p.fontMetrics() ;
    int w = fm.width(label)/2 ;
    label_rect->setCoords (x()+width()/2-w, y() + height(), x() + width()/2+w,
			  y() + height() +p.fontMetrics().lineSpacing()) ;
    p.drawText (*label_rect, AlignCenter, label) ;
    //    p.drawText (x()-xd, y() + height(), xdd, p.fontMetrics().lineSpacing(),
    //	AlignCenter, file->name() ) ;
  }
} 


void Icon::construct_image (QPixmap *pm, IconProgram::iconState state)
{
  QPainter p ;
  p.begin (pm) ;
  if (state == IconProgram::bitmap)
    pm->fill (color0) ;
  else
    pm->fill (QColor (120, 171, 172) ) ;
  p.scale (scale/96.0, scale/96.0) ;
  program->draw (p, state) ;
  p.end () ;   
}

void Icon::rescale (int s)
{
  /*
   * configure the icon for the new scale, 
   * redraw pixmaps if necessary
   */

  s *= 8 ; // rescale slider value to pixels
  if (s <16) s = 16 ;
  if (s > 512) s = 512 ; // our own bounds

  scale = s ;
  bound_rect.setWidth(s) ; bound_rect.setHeight(s) ;

  int n = strlen (icon_file) ;  
  if (n >250) {
    debug ("buffer overflow in Icon::rescale") ;
    exit (1) ;
  }
  
  // now make the mask

  strncpy (mask, icon_file, n) ;
  strncpy (mask+n, "msk", 3) ;
  sprintf (mask+n+3, "%d", s) ; // relies on fact that max s=500
  
  // normal pixmap 
  strncpy (pm_normal, icon_file, n) ;
  strncpy (pm_normal+n, "nor", 3) ;
  sprintf (pm_normal+n+3, "%d", s) ;

  // selected pixmap
  strncpy (pm_selected, icon_file, n) ;
  strncpy (pm_selected+n, "sel", 3) ;
  sprintf (pm_selected+n+3, "%d", s) ;
  
  // focus pixmap 
  strncpy (pm_focus, icon_file, n) ;
  strncpy (pm_focus+n, "foc", 3) ;
  sprintf (pm_focus+n+3, "%d", s) ;

  // drag drop pixmap
  strncpy (pm_drag_drop, icon_file, n) ;
  strncpy (pm_drag_drop+n, "dnd", 3) ;
  sprintf (pm_drag_drop+n+3, "%d", s) ;

  // embossed pixmap
  strncpy (pm_embossed, icon_file, n) ;
  strncpy (pm_embossed+n, "emb", 3) ;
  sprintf (pm_embossed+n+3, "%d", s) ;

}  

void Icon::draw (Display *d, QWidget *desktop)
{
  (void) d , (void) desktop ;
  /*
  HANDLE h1 = pmNormal->handle() ;
  GC gc = XCreateGC (d, h1, 0, 0) ;
  XSetFillStyle (d, gc, FillSolid) ;
  HANDLE h2 = desktop->handle () ;
  XFillRectangle (d, h2, gc, 0, 0, 10, 20) ;
  */
  /*
  XCopyArea (d, pmNormal->handle(), h1,
	     0, 0, w, h, 20, 20) ; */
}






char *Icon::suggest_zic (const char *str) 
{

  /* given a filename, suggest an appropriate zicon
   * file to use as the icon picture. Returns the
   * full path name to the .zic file. This routine
   * should always return a usable file name
   */

  /* 
   * structure of fields:
   * type, name , permission, contents, zicon, openaction dragdrop action
   */
  char *name ;
  QFile f(str) ;
  if (f.exists()) {
    QFileInfo fi (f) ;
    name = match_to_rules (&fi, zicon) ;
  } else {
    name = match_to_rules (str, zicon) ;
  }
  if (name) {
    // ok, add on the right place
    int n = strlen(name)+strlen(ZICDIR)+2 ;
    char *fullname = new char[n] ;
    CHECK_PTR (fullname) ;
    strcpy (fullname, ZICDIR) ;
    strcat (fullname, "/") ;
    strcat (fullname, name) ;
    fullname[n-1] = '\0' ;
    delete[] name ;
    return fullname ;
  }
  // use a default -- should never happen with proper .jafmrules
  debug ("No def zicon") ;
  return "/No/Default/Zicon" ;
}
      
bool Icon::is_drag_droppable () const
{
  /* should we allow drag dropping onto the file f
   */
  char *c ; 
  if (has_file ()) {
    if (finfo->isDir())
      return finfo->isWritable () ;
    
    c = match_to_rules (finfo, dragdropaction) ;
  } else {
    c = match_to_rules (get_label(), dragdropaction) ;
  }
  if (c) {
    if (*c == '-') {
      return false ;
    }
    return true ;
  }
  return false ;
}

void Icon::load_rules ()
{
  if (rule_list) return ; // already loaded!
  rule_list = new QList<QString> ;
  

  QString *str ;
  
  char buf[200] ;
  strcpy (buf, JAFMDIR) ;
  strcat (buf, "/system.jafmrules") ;
  QFile f (buf) ; ;

  if (!f.open (IO_ReadOnly)) {
    debug ("Couldn't open rules file %s",f.name()) ;
    exit (1) ;
  }
  QTextStream stream (&f) ;

  while (!stream.eof()) {
    str = new QString(stream.readLine().simplifyWhiteSpace ()) ;
    if (*str == " " || *((const char *)(*str)) == '#'
	|| str->isEmpty()) {
      delete str ;
      continue ; // comment or empty line
    } 
    rule_list->append (str) ;
  }
}

char *Icon::match_to_rules (const char *name, need_type ntype)
{
  /*
   * check if the name matches our rules
   * 
   */


  QListIterator<QString> it (*rule_list) ;
  QRegExp re ;
  char buffer[200] ; //bug
  char *buf = buffer ;
  int index ;
  re.setWildcard (true) ;

  for ( ; it.current() ; ++it) {
    
    int n = it.current()->length() ;
    if (n == 0) {
      debug ("Zero length rule!") ;
      continue ; 
    }

    strncpy (buf, *(it.current()), n) ;
    
    /*
     * ignore file type
     */

    
    // skip to next field - actually we don't assume it is 1 character long
    index = 1 ;
    while (*(buf+index) != ' ')
      index++ ;
    buf += index + 1 ;
    
    /* Next we match filename
     * we are at filename, get it's length
     * 
     */
    index = 1 ;
    while (*(buf+index) != ' ')
      index++ ;
    
    // now match filename    
    // if it is '*', always match
    if (!(index == 1 && *buf == '*')) {

      *(buf+index) = '\0' ;      
      re = QRegExp (buf) ;
      re.setWildcard (true) ;
      if (re.match (name) < 0)
	continue ;
    }
    
    buf += index + 1 ;
   
    /*
     * now we match the permissions
     */
    
    
    // we ignore permissions for now
    index = 1 ;
    while (*(buf+index) != ' ')
      index++ ;
		
    buf += index + 1 ;
    
    /*
     * Now we match the content
     */

    
    // we ignore contents for now
    index = 1 ;
    while (*(buf+index) != ' ')
      index++ ;
    
    buf += index + 1 ;
    

    /*
     * if we got this far, we have a match
     */
    
    // the order of the next fields are zicon, openaction, dragdropaction
    // the need type is respectively 1 , 2 , 3
    // a ';' separates the open and drag fields, since spaces are significant there
    char matchchar = ' ' ; 
    index = 0 ;
    for (int i = 0 ; i< ntype ; ++i) {
      buf += index ;
      if (i>0) buf++ ; //kludge
      index = 1 ;
      while (*(buf+index) != matchchar)
	index++ ;
      matchchar = ';' ;
    }
    // okay, we are up to the required field, make it a proper string
    if (*buf == ' ') {
      buf++ ; index-- ;
    }
    *(buf+index) = '\0' ;
    
    index++ ;
    char *field = new char[index] ;
    CHECK_PTR (field) ;
    strncpy (field, buf, index) ;
    return field ;

  }
  // nothing matched
  return 0 ;
}


char * Icon::match_to_rules (QFileInfo *fi, need_type ntype)
{

  /* check if the file matches our rules
   * return the requested information if it does
   */

  QListIterator<QString> it (*rule_list) ;
  QRegExp re ;
  char buffer[200] ;
  char *buf = buffer ;
  int index ;
  re.setWildcard (true) ;
  /*  QDir save_dir = QDir::current () ;
  QDir::setCurrent (fi->dirPath (true)) ; // ignoring return... */

  for ( ; it.current() ; ++it) {
    
    int n = it.current()->length() ;
    strncpy (buf, *(it.current()), n) ;
    
    /*
     * first thing we do is match the type
     */

    if (! (
	   ( *buf == '*')
	   || (  *buf == 'D' && fi->isDir())
	   || *buf == 'F' && fi->isFile())) continue ;
    
    // skip to next field - actually we don't assume it is 1 character long
    index = 1 ;
    while (*(buf+index) != ' ')
      index++ ;
    buf += index + 1 ;
    
    /* Next we match filename
     * we are at filename, get it's length
     * 
     */
    index = 1 ;
    while (*(buf+index) != ' ')
      index++ ;
    
    // now match filename    
    // if it is '*', always match
    if (!(index == 1 && *buf == '*')) {

      *(buf+index) = '\0' ;      
      re = QRegExp (buf) ;
      re.setWildcard (true) ;
      if (re.match (fi->fileName()) < 0)
	continue ;
    }
    
    buf += index + 1 ;
   
    /*
     * now we match the permissions
     */

    index = 0 ;
    bool failed = false ;
    if (*buf != '*') {
      while (*(buf+index) != ' ') {
	switch (*(buf+index)) {
	  
	case 'x': // can exec
	  if (!(fi->isExecutable()))
	    failed = true ;
	  break ;
	default:
	  debug ("Unknown permission") ;
	}
      index++ ;
      }
    }
    if (failed) continue ;
    buf += index + 1 ;
    
    /*
     * Now we match the content
     */

    
    // we ignore contents for now
    index = 1 ;
    while (*(buf+index) != ' ')
      index++ ;
    
    buf += index + 1 ;
    

    /*
     * if we got this far, we have a match
     */
    
    // the order of the next fields are zicon, openaction, dragdropaction
    // the need type is respectively 1 , 2 , 3
    // a ';' separates the open and drag fields, since spaces are significant there
    char matchchar = ' ' ; 
    index = 0 ;
    for (int i = 0 ; i< ntype ; ++i) {
      buf += index ;
      if (i>0) buf++ ; //kludge
      index = 1 ;
      while (*(buf+index) != matchchar)
	index++ ;
      matchchar = ';' ;
    }
    // okay, we are up to the required field, make it a proper string
    if (*buf == ' ') {
      buf++ ; index-- ;
    }
    *(buf+index) = '\0' ;
    
    index++ ;
    char *field = new char[index] ;
    strncpy (field, buf, index) ;
    // QDir::setCurrent (save_dir.path()) ;
    return field ;


  }
  // nothing matched
  //  QDir::setCurrent (save_dir.path()) ;
  return 0 ;
}

void Icon::drag_drop_action (QList<DeskIcon> &ic_list)
{
  /* 
   * a bunch of icons - icList -
   * have been dragged onto us
   */
  
  // Are we a directory?
  if (is_dir()) {
    debug ("I'm a directory") ;
    QListIterator<DeskIcon> it(ic_list) ;
    for (; it.current() ; ++it) {
      if (it.current()->get_icon()->has_file()) {
	// make the new filename
	int n = strlen (get_full_name()) + strlen (it.current()->get_icon()->get_label()) + 2 ;
	char *name = new char[n] ;
	strcpy (name,get_full_name()) ;
	strcat (name, "/") ;
	strcat (name, it.current()->get_icon()->get_label()) ;
	debug ("Move %s",(const char *) ic_list.getFirst()->get_icon()->get_label()) ;
	debug ("To %s",  name) ;
	if (rename (ic_list.getFirst()->get_icon()->get_label(), name)) {
	  perror ("rename") ;
	}
	delete name ;
      } else {
	debug ("%s is not a real file - can't move",  it.current()->get_icon()->get_label()) ;
      }
    }
    return ;
  }
  
  QString execstr ;
  int p ;

  if (has_file()) {
    execstr = Icon::match_to_rules (finfo, Icon::dragdropaction) ;
  } else {
    const char *lab = get_label() ;
    execstr = match_to_rules (lab, dragdropaction) ;
  }
  
  // a dash means we shouldn't have even got this far!
  if (*((const char *)(execstr)) != '-') {

    char *argList[200] ; // assume no more than 200 args!
    
    // a ! means use the default 
    if ( *((const char *)(execstr)) == '!') {
      argList[0] = new char[strlen (label)+1] ;
      strcpy (argList[0],get_label()) ;
  
      // copy in the file names as arguments
      uint i ;
      for (i = 0 ; i < ic_list.count() ; ++i) {
	int n = strlen (ic_list.at(i)->get_icon()->get_label()) ;
	argList[i+1] = new char[n+1] ;
	strcpy (argList[i+1], ic_list.at(i)->get_icon()->get_label()) ;
      }
      argList[i+1] = 0 ;
      p = i ;
    } else {
      
      // ok, parse it
      // replace % character with the label
      QListIterator<DeskIcon> it (ic_list) ;
      QString labels ;
      bool notfirst = false ;
      for ( ; it.current() ; ++it) {
	if (notfirst) labels += ' ' ;
	notfirst = true ;
	labels += it.current()->get_icon()->get_label() ;
      }
      debug ("dragged dropped %s", (const char *) labels) ;
      execstr.replace (QRegExp ("%"), labels) ;
      // divide up into an arg list
      
      execstr.simplifyWhiteSpace () ; // buggy
      
      uint jj = 0 ;
      uint j  ;
      p = 0 ;
      for (j = 0 ; j < execstr.length (); ++j) {
	if (*(((const char*)execstr)+j) == ' ') {
	  argList[p] = new char[j-jj+1] ;
	  strncpy (argList[p], execstr.mid(jj,j),j-jj) ;
	  argList[p++][j-jj] = '\0' ;
	  jj = j+1 ; // +1 ignores the space
	}
      }
      // do last one 
      if (j-jj != 0) {
	argList[p] = new char[j-jj+1] ;
	strncpy (argList[p], execstr.mid(jj,j),j-jj) ;
	argList[p++][j-jj] = '\0' ;  
      }
      argList[p] = 0 ;
    }
    if (!fork ()) { //child
	
      // build up an arg list
      
      if (!fork()) {
	execvp (argList[0],argList ) ;
	perror ("exec failed") ;
	exit (1) ;
      }
      exit (0) ;
    }
    for (int j = 0; j < p ; ++j) delete[] argList[j] ;
  }
}

void Icon::open_action ()
{
  /*
   * we have been asked to open ourselves,
   * do whatever seems appropritate
   */
  
  // if we are a directory, we should never have been called
  
  
  // get the string telling us what to do
  QString execstr ;
  if (has_file()) {
    execstr = Icon::match_to_rules (finfo, Icon::openaction) ;
  } else {
    const char *lab = get_label() ;
    execstr = match_to_rules (lab, openaction) ;
  }
  
  if (!execstr) {
    debug ("Zero execstr: error in .jafmrules?") ;
    exit (1) ;
  }
  

  // a dash means it shouldn't be opened
  if (*((const char *)(execstr)) != '-') {

    char *argList[200] ; // limit of 200 args
    int p = 0 ;
    
    // a ! means use the default action: just execute the file
    if ( *((const char *)(execstr)) == '!') {
      argList[0] = new char[strlen (label)+1] ;
      strcpy (argList[0], label) ;
      argList[1] = 0 ;
      p++ ;
    } else {
    
      // ok, parse it
      // replace % character with the label
      execstr.replace (QRegExp ("%"), get_label ()) ;
      
      // divide up into an arg list
      execstr.simplifyWhiteSpace () ; // buggy?
      
      uint jj = 0 ;
      uint j  ;
      for (j = 0 ; j < execstr.length (); ++j) {
	if (*(((const char*)execstr)+j) == ' ') {
	  argList[p] = new char[j-jj+1] ;
	  strncpy (argList[p], execstr.mid(jj,j),j-jj) ;
	  argList[p++][j-jj] = '\0' ;
	  jj = j+1 ; // +1 ignores the space
	}
      }
      // do last one 
      argList[p] = new char[j-jj+1] ;
      strncpy (argList[p], execstr.mid(jj,j),j-jj) ;
      argList[p++][j-jj] = '\0' ;  
      argList[p] = 0 ;
    }

    if (argList[0][0] == '/') {
      QFile f (argList[0]) ;
      QFileInfo fi (f) ;
      if (fi.isDir()) {
	debug ("open was dir") ;
	emit (open_dir_window (argList[0])) ;
	return ;
      }
    }
	

    // special case: if the 'program' to be executed is
    // a directory (on its own) then open it in new window

    if (!fork ()) { //child
      
      // build up an arg list
      
      if (!fork()) {
	execvp (argList[0],argList ) ;
	perror ("exec failed") ;
	exit (1) ;
      }
      exit (0) ;
    }
    for (int i = 0 ; i <p ; ++i) delete[] argList[i] ;
  }
}



QBitmap * Icon::get_mask ()
{
  QBitmap *bm = (QBitmap *) QPixmapCache::find (mask) ;
  if (!bm) {
    bm = new QBitmap (scale, scale) ;
    construct_image (bm, IconProgram::bitmap) ;
    QPixmapCache::insert (mask, bm) ;
  }
  return bm ;
}

QPixmap * Icon::get_normal ()
{
  QPixmap *pm = QPixmapCache::find (pm_normal) ;
  if (!pm) {
    pm = new QPixmap (scale, scale) ;
    construct_image (pm, IconProgram::normal) ;
    pm->setMask (*get_mask()) ;
    QPixmapCache::insert (pm_normal, pm) ;
  }
  return pm ;
}

QPixmap * Icon::get_embossed ()
{
  QPixmap *pm = QPixmapCache::find (pm_embossed) ;
  if (!pm) {
    pm = new QPixmap (scale, scale) ;
    construct_image (pm, IconProgram::emboss) ;
    pm->setMask (*get_mask()) ;
    QPixmapCache::insert (pm_embossed, pm) ;
  }
  return pm ;
}

QPixmap * Icon::get_focus ()
{
  QPixmap *pm = QPixmapCache::find (pm_focus) ;
  if (!pm) {
    pm = new QPixmap (scale, scale) ;
    construct_image (pm, IconProgram::focus) ;
    pm->setMask (*get_mask()) ;
    QPixmapCache::insert (pm_focus, pm) ;
  }
  return pm ;
}

QPixmap * Icon::get_selected ()
{
  QPixmap *pm = QPixmapCache::find (pm_selected) ;
  if (!pm) {
    pm = new QPixmap (scale, scale) ;
    construct_image (pm, IconProgram::selected) ;
    pm->setMask (*get_mask()) ;
    QPixmapCache::insert (pm_selected, pm) ;
  }
  return pm ;
}

QPixmap * Icon::get_drag_drop ()
{
  QPixmap *pm = QPixmapCache::find (pm_drag_drop) ;
  if (!pm) {
    pm = new QPixmap (scale, scale) ;
    construct_image (pm, IconProgram::dragdrop) ;
    pm->setMask (*get_mask()) ;
    QPixmapCache::insert (pm_drag_drop, pm) ;
  }
  return pm ;
}

void Icon::new_file_info (QFileInfo  &f) 
{
  delete finfo ;
  finfo = new QFileInfo (f) ;
  delete [] label ;
  label = new char[strlen (finfo->fileName()) + 1] ;
  strcpy (label,finfo->fileName() ) ;
  //  update icon
    
}

/*
 * icon program routines
 */

extern FILE* yyin ;
extern int yyparse (void *) ;

extern "C" void yyrestart (FILE *) ;


IconProgram::IconProgram (const char *name) { 
  progSize = 1000 ; //default
  program = new short int[progSize] ;
  instructions = 0 ;
  realSize = 0 ;

  static bug = 0 ;
  
  // okay, parse the file
  // error checking is s.e.p.
  FILE *file = fopen (name, "r") ;
  if (!file) {
    perror ("fopen") ;
    cerr << "Fatal: Couldn't open the zicon file \"" 
	 << name << "\" even though I had been assured I could!\n" ;
    exit (1) ;
  }
  
  if (bug == 1) {
    yyrestart (file) ;
  } else {
    yyin = file ;
    bug = 1 ;
  }
  
  yyparse (this) ;
  fclose (file) ;
  
}

void IconProgram::draw (QPainter &p,IconProgram::iconState state ) 
{

  int pos = 0 ;
  QPointArray q ;  
  // get the instruction type, then execute:

  for (int i = 0 ; i < instructions ; ++i) {

    switch (program[pos++]) {
      
    case pen_iconcolour:

      switch (state) {
	
      case normal:
	p.setPen (QColor (210, 210, 210)) ;
	break ;

      case selected:
	p.setPen (QColor (220, 220, 0)) ;
	break ;
	
      case focus:
	p.setPen (QColor (255,255,255)) ;
	break ;
	
      case dragdrop:
	p.setPen (QColor (80, 80, 200)) ;
	break ;
	
      case bitmap:
      case emboss:
	break ;
      }
      break ;

    case brush_iconcolour:

      switch (state) {
	
      case normal:
	p.setBrush (QColor (210, 210, 210)) ;
	break ;

      case selected:
	p.setBrush (QColor (220, 220, 0)) ;
	break ;
	
      case focus:
	p.setBrush (QColor (255, 255, 255)) ;
	break ;
	
      case dragdrop:
	p.setBrush (QColor (80, 80, 200)) ;
	break ;
	
      case bitmap:
      case emboss:
	break ;
      }
      break ;

    case pen_shadowcolour:
      p.setPen (QColor(60, 60, 60)) ;
      break ;
      
    case brush_shadowcolour:
      p.setBrush (QColor (60, 60, 60)) ;
      break ;

    case pen_outlinecolour:
      p.setPen (black) ;
      break ;

    case brush_outlinecolour:
      p.setPen (black) ;
      break ;
      
    case pen_colour:
      p.setPen (QColor (program[pos], program[pos+1], program[pos+2])) ;
      pos += 3 ;
      break ;

    case brush_colour:
      p.setBrush (QColor (program[pos], program[pos+1], program[pos+2])) ;
      pos += 3 ;
      break ;
      
    case fillpoly:
      
      if (state == bitmap) {
	p.setPen (color1) ;
	p.setBrush (color1) ;
      }
      if (state == emboss) {
	p.setBrush ((QColor (90, 90, 90))) ;
	p.setPen ((QColor (90, 90, 90))) ;
	// p.setRasterOp (NotCopyROP) ;
      }
      q.setPoints (program[pos]/2, &program[pos+1]); 
      p.drawPolygon (QPointArray (q)) ;
      pos += program[pos] + 1 ;
      break ;

    case linepoly:
      if (state == bitmap) {
	p.setPen (color1) ;
	p.setBrush (color1) ;
      }
      q.setPoints (program[pos]/2, &program[pos+1]); 
      p.drawPolyline (QPointArray (q)) ;
      pos += program[pos] + 1 ;
      break ; 

    default:
      debug ("error in icon program code") ;
    } ;
  }
}


void IconProgram::addInstruction (const short int *code, const int n)
{
  if (n + realSize > progSize) {
    debug ("Need more space for icon program") ;
  }
  for (int j = 0 ; j < n ; ++j)
    program[realSize + j] = code[j] ;
  realSize += n ;
  instructions++ ;
}  

void IconProgram::addInstruction (const short int i) 
{
  if ( 1 + realSize > progSize) {
    debug ("Need more space for icon program") ;
      
  }
  program[realSize] = i ;
  instructions++ ;
  realSize++ ;
}
